#include "StateManager.h"
#include "ESP8266WiFi.h"

StateManager::StateManager() {
  m_debug = false;
  m_lastKVPUpdate = 0;
  m_roolOverIsPossible = false;
  m_uptimeDays = 0;
}

void StateManager::SetDebugMode(boolean mode) {
  m_debug = mode;
}

void StateManager::Begin(String version) {
  m_version = version;
  m_settings.Read();
}

void StateManager::Handle(byte frames){
  m_receivedFrames += frames;
  unsigned long currentMillis = millis();
  
  // Handle 50 Days rollover for UpTime
  if(currentMillis >= 3000000000) { 
    m_roolOverIsPossible = true;
  }
  if(currentMillis <= 100000 && m_roolOverIsPossible) {
    m_uptimeDays += 50;
    m_roolOverIsPossible = false;
  }
    
  // Calculate frames per minute
  m_framesPerMinute = 0;
  for(int i=0; i<TIMES_SIZE; i++){
    if(currentMillis > 60000 && m_times[i] < currentMillis -60000) {
      m_times[i] = 0;
    }
    
    if(frames > 0) {
      if(m_times[i] == 0) {
        m_times[i] = currentMillis;
        frames--;
      }
    }
    
    if(m_times[i] > 0) {
      m_framesPerMinute++;
    }
    
  }
}

String StateManager::GetVersion() {
  return m_version;
}

void StateManager::Update() {
  m_values.Clear();

  unsigned long upTimeSeconds = millis() / 1000;
  String upTimeText(m_uptimeDays + upTimeSeconds / 86400);
  upTimeText += "Tg. ";
  upTimeText += String(upTimeSeconds / 3600 % 24);
  upTimeText += "Std. ";
  upTimeText += String((upTimeSeconds / 60) % 60);
  upTimeText += "Min. ";
  upTimeText += String(upTimeSeconds % 60);
  upTimeText += "Sek. ";

  
  m_values.Put("UpTimeSeconds", String(upTimeSeconds));
  m_values.Put("UpTimeText", upTimeText);
  m_values.Put("WIFI", WiFi.SSID());
  m_values.Put("MacAddress", WiFi.macAddress());
  m_values.Put("ChipID", String(ESP.getChipId()));
  m_values.Put("ReceivedFrames", String(m_receivedFrames));
  m_values.Put("FramesPerMinute", String(m_framesPerMinute));
  m_values.Put("RSSI", String(WiFi.RSSI()));


}

String StateManager::GetHTML() {
  Update();
  String result = "<style>table{border:1px solid black;border-collapse:collapse;}tr,td{padding:4px;border:1px solid; Margin:5px}</style>";
  
  result += "<table>";
  for (int i = 0; i < m_values.Size(); i++) {
    result += "<tr><td>";
    result += m_values.GetKeyAt(i);
    result += ": </td><td>";
    result += m_values.GetValueAt(i);
    result += "</td></tr>";
  }
  result += "</table>";

  return result;
}

String StateManager::GetXML() {
  Update();
  String result = "";

  result += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
  result += "<LGW>";
  for (int i = 0; i < m_values.Size(); i++) {
    result += "<Info Key = \"";
    result += m_values.GetKeyAt(i);
    result += "\" Value=\"";
    result += m_values.GetValueAt(i);
    result += "\"/>";
  }
  result += "</LGW>";
  
  return result;
}

String StateManager::GetKVP() {
  Update();
  String result = "";

  result += "OK VALUES LGW ";
  result += m_settings.Get("KVIdentity", String(ESP.getChipId()));

  result += " ";

  for (int i = 0; i < m_values.Size(); i++) {
    result += m_values.GetKeyAt(i);
    result += "=";
    result += m_values.GetValueAt(i);
    if (i != m_values.Size() -1) {
      result += ",";
    }
  }

  return result;
}

String StateManager::GetKVP(uint interval) {
  String result = "";
  
  if (millis() < m_lastKVPUpdate) {
    m_lastKVPUpdate = 0;
  }
  if (millis() > m_lastKVPUpdate + interval * 1000) {
    result = GetKVP();
    m_lastKVPUpdate = millis();
  }

  return result;
}

